/**
*   Please feel free to use any fragment of the code in this file that you need
*   in your own work. As far as I am concerned, it's in the public domain. No
*   permission is necessary or required. Credit is always appreciated if you
*   use a large chunk or base a significant product on one of my examples,
*   but that's not required either.
*
*   This code is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*      --- Joseph A. Huwaldt
**/

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;

import net.java.games.jogl.*;

import jahuwaldt.gl.*;


/**
*  A class renders a 3D terrain into an OpenGL context supplied by JOGL.
*
*  <p>  Modified by:  Joseph A. Huwaldt  </p>
*
*  @author  Joseph A. Huwaldt   Date:  April 19, 2001
*  @version November 18, 2004
**/
public class LandscapeRenderer implements GLEventListener, MouseListener, MouseMotionListener, KeyListener {

	//  Path (relative to directory where Java VM is started) to map data file.
	private static final String kDataFilePath = "data/Height1024.raw";
	
	//	Map scale.
	private static final float kAltScale = 1;
	private static final float kGridSpacing = 2;
	
	//	For perspective calculation.
	private static final float kNearClip = 1;
	private static final float kFarClip = 1000*kGridSpacing;

	//	The width and height of the altitude map.
	private static int kMapXSize = 1024;
	private static int kMapYSize = 1024;
	
	//	The patch size to use in the GeoMM algorithm.
	private static final int kPatchSize = 64;
	
	//	The height map data.
	private HeightMap heightMap;
	
	//	The size of the texture map.
	private static final int kTextureSize = 128;
	
	//	The texture ID.
	private int gTextureID = 1;
	
	// Camera modes
	public static final int FOLLOW_MODE = 0;
	public static final int OBSERVE_MODE = 1;
	public static final int DRIVE_MODE = 2;
	public static final int FLY_MODE = 3;
	
	// Rotation Indexes
	private static final int ROTATE_PITCH = 0;
	private static final int ROTATE_YAW = 1;
	private static final int ROTATE_ROLL = 2;

	//	The current camera mode.
	private int gCameraMode = FOLLOW_MODE;
	
	//	Camera view position in model coordinates in follow, drive and fly modes.
	private float gViewPosition[] = { 0, 0, 0 };
	
	//	Camera position in model coordinates when in observe mode.
	private float gCameraPosition[] = { 0.f, 0.f, -555.f*kGridSpacing };
	
	//	Camera rotation.
	private float gCameraRotation[] = { 42.f, -181.f, 0.f };

	//	Camera field of view in degrees.
	private float gFovX = 60.0f;
	
	//	Current rotation angle for animation.
	private float gAnimateAngle = 0;
	
	//	The angle the view is centered on.  Used for frustum culling.
	private float gClipAngle;
	
	//	Misc. flags.
	private boolean gAnimating = (gCameraMode == FOLLOW_MODE);
	private boolean gDrawFrustum = true;
	private boolean gDrawFog = true;
	private boolean gUpdateFog = true;			//  Fog needs updating.
	private boolean gUpdateDrawMode = true;		//  Draw mode needs updating.
	
   	//	A reference to our landscape rendering algorithm.
   	private GeoMMLandscape gLand;

	
	/**
	*  Construct a landscape renderer for rendering a height map in 3D.
	**/
	public LandscapeRenderer() throws IOException {
	
		//	Read in the height map from disk.
		heightMap = loadTerrain(kMapXSize, kMapYSize);
		
	   	//	Initialize the GeoMipMap algorithm.
    	gLand = new GeoMMLandscape(heightMap, kPatchSize, kGridSpacing,
    								kNearClip, (float)(kNearClip*Math.tan( gFovX/2.0*Math.PI/180.0 )), 600, 4);
    	gLand.setDrawMode(LODLandscape.kUseTexture);
    	
	}
	
	
	/**
	*  Set the drawing mode to one of the constants defined
	*  in LODLandscape.
	**/
	public void setDrawMode(int mode) {
		if (gLand != null) {
			gUpdateDrawMode = true;
			gLand.setDrawMode(mode);
		}
		
	}
	
	/**
	*  Set the camera mode to one of the constants defined
	*  in this class.
	**/
	public void setCameraMode(int mode) {
		if (mode < 0)
			mode = 0;
		else if (mode > 3)
			mode = 3;
		
		gCameraMode = mode;
		
		if (mode == FOLLOW_MODE) {
			// Turn animation back on if needed.
			gAnimating = true;
		
		} else if (mode == DRIVE_MODE || mode == FLY_MODE) {
			//	Stop animation in drive and fly modes.
			gAnimating = false;
		}
		
		//	Deal with fog.
		gUpdateFog = true;
		if (mode == OBSERVE_MODE)
			gDrawFog = false;
		else
			gDrawFog = true;
		
	}
	
	/**
	*  Indicates if the field of view frustum is drawn or not.
	**/
	public boolean drawFrustum() {
		return gDrawFrustum;
	}
	
	/**
	*  Set to true to draw the field of view frustum, false to
	*  not draw it.
	**/
	public void drawFrustum(boolean flag) {
		gDrawFrustum = flag;
	}
	
	/**
	*  Indicates if fog should be drawn or not.
	**/
	public boolean drawFog() {
		return gDrawFog;
	}
	
	/**
	*  Set to true to render with fog effects, false to
	*  not draw fog.
	**/
	public void drawFog(boolean flag) {
		gDrawFog = flag;
		gUpdateFog = true;
	}
	
	/**
	*  Returns the pixel threshold used when deciding to switch to higher
	*  geo mipmap levels.  Indicates how much detail is being drawn.
	**/
	public int getLevelOfDetail() {
		return gLand.getLevelOfDetail();
	}
	
	/**
	*  Set the pixel threshold used when deciding to switch to higher
	*  geo mipmap levels.  Indicates how much detail is being drawn.
	**/
	public void setLevelOfDetail(int lod) {
		gLand.setLevelOfDetail(lod);
	}
	
	/**
	*  Return the actual number of triangles rendered.
	**/
	public int numTrianglesRendered() {
		return gLand.numTrianglesRendered();
	}


	/**
	*  Initialize this renderer.
	**/
    public void init(GLDrawable drawable) {
		
		//  Get reference to GL context.
		GL gl = drawable.getGL();
		
		System.err.println("INIT GL IS: " + gl.getClass().getName());
		System.err.println("GL_VENDOR: " + gl.glGetString(GL.GL_VENDOR));
		System.err.println("GL_RENDERER: " + gl.glGetString(GL.GL_RENDERER));
		System.err.println("GL_VERSION: " + gl.glGetString(GL.GL_VERSION));

		//  Add event listeners to the drawable object.
		drawable.addMouseListener(this);
		drawable.addMouseMotionListener(this);
		drawable.addKeyListener(this);
		
		//	Setup the rendering contex.
		Dimension size = drawable.getSize();
		setupRenderingContext(gl, size.width, size.height);

		//	Setup the textures used by this canvas.
		setupTextures(gl, drawable.getGLU());
		
    	//	Set the drawing mode context.
		setDrawModeContext(gl, gLand.getDrawMode());
		gUpdateDrawMode = false;
	}
	
	
	/**
	*  Called whenever the drawable changes size.
	**/
	public void reshape(GLDrawable drawable, int x, int y, int width, int height) {
	
		// Prevent a divide by zero, when the canvas is too small.
		if (height == 0)
			height = 1;
		if (width == 0)
			width = 1;

		//  Get reference to GL context.
		GL gl = drawable.getGL();
		
		// Set the viewport to be the entire canvas.
    	gl.glViewport(0, 0, width, height);

		// Set up the perspective matrix.
		setupPerspective(gl, width, height);
	}
	
	
	/**
	*  Called whenever the display needs to be re-rendered.
	**/
	public void display(GLDrawable drawable) {
	
		// If animating, move the view position
		if (gAnimating) {
			gAnimateAngle += 0.4f;

			double angle = gAnimateAngle*Math.PI/180;
			gViewPosition[0] = kGridSpacing*(kMapXSize/4 + (float)(Math.sin(angle) + 1)*kMapXSize/4 );
			gViewPosition[2] = kGridSpacing*(kMapYSize/4 + (float)(Math.cos(angle) + 1)*kMapYSize/4 );
			
			gViewPosition[1] = heightAtViewPosition();
		}
		
		//  Get reference to GL context.
		GL gl = drawable.getGL();
		
		//  If the draw mode has changed, update draw mode context.
		if (gUpdateDrawMode) {
			gUpdateDrawMode = false;
			setDrawModeContext(gl, gLand.getDrawMode());
		}
			
		//  Update fog settings if necessary.
		if (gUpdateFog) {
			gUpdateFog = false;
			setFogFlag(gl, gDrawFog);
		}

		//	Clear the GL buffer.
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		//	Save off the current matrix.
		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glPushMatrix();

		//	Reset the rotation matrix
		gl.glLoadIdentity();

		//	Begin rendering.
		setCameraModeContext(gl);
		
		// Perform the actual rendering of the mesh.
		gLand.update(gFovX, gViewPosition, gClipAngle);
		gLand.render(gl);

		if ( gDrawFrustum )
			doDrawFrustum(gl);

		//	Reset transformation matrix.
		gl.glPopMatrix();
		
	}
	
    public void displayChanged(GLDrawable drawable, boolean modeChanged, boolean deviceChanged) {}

	// ---------------------------------------------------------------------
	/**
	*  Set up the rendering context.
	*  This function does any needed initialization on the rendering
	*  context.  Here it sets up and initializes the lighting for
	*  the scene.
	**/
	private void setupRenderingContext(GL gl, int width, int height) {
	
		//	This will clear the background color to sky blue
		float skyColor[] = {0.75f, 0.75f, 1.0f, 1.0f};
		gl.glClearColor( skyColor[0], skyColor[1], skyColor[2], skyColor[3] );

		//	Enables clearing of the depth buffer
		gl.glClearDepth(1.0);

		//	The type of depth test to do
		gl.glDepthFunc(GL.GL_LESS);
		
		//	Enable Hidden surface removal
		gl.glEnable(GL.GL_DEPTH_TEST);
		
		//	Cull back-facing triangles
		gl.glCullFace(GL.GL_BACK);
		gl.glEnable(GL.GL_CULL_FACE);
		
		//	Set the viewport to be the entire window
    	gl.glViewport(0, 0, width, height);

		//	Set up the perspective matrix.
		setupPerspective(gl, width, height);
		
		
		// ----------- LIGHTING SETUP -----------
		// Light values and coordinates
		float whiteLight[]    = { 0.45f,  0.45f, 0.45f, 1.0f };
		float ambientLight[]  = { 0.25f,  0.25f, 0.25f, 1.0f };
		float diffuseLight[]  = { 0.50f,  0.50f, 0.50f, 1.0f };
		float lightPos[]      = { 0.00f,300.00f, 0.00f, 0.0f };
	
		// Setup and enable light 0
		gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, ambientLight);
		gl.glLightfv(GL.GL_LIGHT0,GL.GL_AMBIENT, ambientLight);
		gl.glLightfv(GL.GL_LIGHT0,GL.GL_DIFFUSE, diffuseLight);
		gl.glLightfv(GL.GL_LIGHT0,GL.GL_POSITION, lightPos);
		gl.glEnable(GL.GL_LIGHT0);

		// Enable color tracking
		gl.glEnable(GL.GL_COLOR_MATERIAL);
	
		// Set material properties to follow color values
		gl.glColorMaterial(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE);

		// Set the color for the landscape
		gl.glMaterialfv( GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, whiteLight );


		//	Enable and set up fog.
		gl.glEnable(GL.GL_FOG);
		gl.glFogi(GL.GL_FOG_MODE,GL.GL_LINEAR);
		gl.glFogfv(GL.GL_FOG_COLOR,skyColor);
		gl.glFogf(GL.GL_FOG_DENSITY,1);
		gl.glFogf(GL.GL_FOG_START,64*kGridSpacing);
		gl.glFogf(GL.GL_FOG_END,512*kGridSpacing);
		
	}


	/**
	*  Set up the perspective transformation matrix.
	**/
	private void setupPerspective(GL gl, int w, int h) {
	
		// Select The Projection Matrix
		gl.glMatrixMode(GL.GL_PROJECTION);
		
		// Reset the perspective matrix
		gl.glLoadIdentity();
		
		// Produce the perspective projection
		double fAspect = (double)w/(double)h;
		double right = kNearClip*Math.tan( gFovX/2.0*Math.PI/180.0 );
		double top = right/fAspect;
		double bottom = -top;
		double left = -right;
		gl.glFrustum(left, right, bottom, top, kNearClip, kFarClip);
		
		// Select The Modelview Matrix
		gl.glMatrixMode(GL.GL_MODELVIEW);		
		gl.glLoadIdentity();
		
	}
	
	/**
	*  Setup any textures used by this program.
	**/
	private void setupTextures(GL gl, GLU glu) {
	
		//	Texture initialization.
		gl.glBindTexture(GL.GL_TEXTURE_2D, gTextureID);
		
		//	Allocate memory for the texture map.
		byte[] pTexture = new byte[kTextureSize*kTextureSize*3];
		
		// Create a random stipple pattern for the texture.  Only use the Green channel.
		// This could easily be modified to load in a bitmap or other texture source.
		int pos = 0;
		for ( int x = 0; x < kTextureSize; x++ ) {
			for ( int y = 0; y < kTextureSize; y++ ) {
				int color = (int)(128.0 + 40.0*Math.random() );
				if ( color > 255 )  color = 255;
				if ( color < 0 )    color = 0;
				
				pTexture[pos++] = (byte)0;
				pTexture[pos++] = (byte)color;	// Only use the Green chanel.
				pTexture[pos++] = (byte)0;
			}
		}
		
		glu.gluBuild2DMipmaps(GL.GL_TEXTURE_2D, 3, kTextureSize, kTextureSize,
											GL.GL_RGB, GL.GL_UNSIGNED_BYTE, pTexture);
    	gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
    	gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST_MIPMAP_NEAREST);
    	
    	gl.glTexEnvi( GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE );
    

		// Use Generated Texture Coordinates
		float s_vector[] = { 1/(float)kTextureSize, 0, 0, 0 };
		float t_vector[] = { 0, 0, 1/(float)kTextureSize, 0 };

		gl.glTexGeni( GL.GL_S, GL.GL_TEXTURE_GEN_MODE, GL.GL_OBJECT_LINEAR );
		gl.glTexGenfv( GL.GL_S, GL.GL_OBJECT_PLANE, s_vector );

		gl.glTexGeni( GL.GL_T, GL.GL_TEXTURE_GEN_MODE, GL.GL_OBJECT_LINEAR );
		gl.glTexGenfv( GL.GL_T, GL.GL_OBJECT_PLANE, t_vector );
		
    }

	/**
	*  Set OpenGL Fog flag.
	**/
	private static void setFogFlag(GL gl, boolean flag) {
		if (flag)
			gl.glEnable(GL.GL_FOG);
		else
			gl.glDisable(GL.GL_FOG);
	}
	
	/**
	*  Switch GL Contexts when moving between draw modes to improve performance.
	**/
	private static void setDrawModeContext(GL gl, int drawMode) {
		switch (drawMode) {
			case LODLandscape.kUseTexture:
				gl.glDisable(GL.GL_LIGHTING);
				gl.glEnable(GL.GL_TEXTURE_2D);
				gl.glEnable(GL.GL_TEXTURE_GEN_S);
				gl.glEnable(GL.GL_TEXTURE_GEN_T);
				gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
				break;

			case LODLandscape.kUseFlatShading:
				gl.glEnable(GL.GL_LIGHTING);
				gl.glDisable(GL.GL_TEXTURE_2D);
				gl.glDisable(GL.GL_TEXTURE_GEN_S);
				gl.glDisable(GL.GL_TEXTURE_GEN_T);
				gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
				break;

			case LODLandscape.kUseGauradShading:
				gl.glDisable(GL.GL_LIGHTING);
				gl.glDisable(GL.GL_TEXTURE_2D);
				gl.glDisable(GL.GL_TEXTURE_GEN_S);
				gl.glDisable(GL.GL_TEXTURE_GEN_T);
				gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
				break;

			case LODLandscape.kUseWireFrame:
			default:
				gl.glDisable(GL.GL_LIGHTING);
				gl.glDisable(GL.GL_TEXTURE_2D);
				gl.glDisable(GL.GL_TEXTURE_GEN_S);
				gl.glDisable(GL.GL_TEXTURE_GEN_T);
				gl.glPolygonMode(GL.GL_FRONT, GL.GL_LINE);
				break;
		}
	}

	/**
	*  Adjusts the rendering context based on the current camera mode.
	**/
	private void setCameraModeContext(GL gl) {
	
		switch (gCameraMode) {
		default:
		case FOLLOW_MODE:
			// Face the center of the map, while walking in a circle around the midpoint (animated).
			gl.glRotatef(-gAnimateAngle, 0.f, 1.f, 0.f);
			gl.glTranslatef(-gViewPosition[0], -gViewPosition[1], -gViewPosition[2]);

			gClipAngle = -gAnimateAngle;
			break;

		case OBSERVE_MODE:
			// Face the center of the map from a certain distance & allow user to orbit around the midpoint.

			// Move to the Camera Position
			gl.glTranslatef(0.f, gCameraPosition[1], gCameraPosition[2]);

			// Rotate to the Camera Rotation angle
			gl.glRotatef(gCameraRotation[ROTATE_PITCH], 1.f, 0.f, 0.f);
			gl.glRotatef(gCameraRotation[ROTATE_YAW],   0.f, 1.f, 0.f);

			// Adjust the origin to be the center of the map...
			gl.glTranslatef(-kMapXSize*0.5f*kGridSpacing, 0.f, -kMapYSize*0.5f*kGridSpacing);
		
			gClipAngle = -gAnimateAngle;
			break;

		case DRIVE_MODE:
		case FLY_MODE:
			// Rotate to the Camera Rotation angle
			gl.glRotatef(gCameraRotation[ROTATE_PITCH], 1.f, 0.f, 0.f);
			gl.glRotatef(gCameraRotation[ROTATE_YAW],   0.f, 1.f, 0.f);

			// Move to the Camera Position
			gl.glTranslatef(-gViewPosition[0], -gViewPosition[1], -gViewPosition[2]);

			gClipAngle = gCameraRotation[ROTATE_YAW];
			gAnimateAngle = -gClipAngle;
			break;
		}

	}

	/**
	*  Draw a simplistic frustum for debug purposes.
	**/
	private void doDrawFrustum(GL gl) {
		//
		// Draw the camera eye & frustum
		//
		gl.glDisable(GL.GL_LIGHTING);
		gl.glDisable(GL.GL_TEXTURE_2D);
		gl.glDisable(GL.GL_TEXTURE_GEN_S);
		gl.glDisable(GL.GL_TEXTURE_GEN_T);

		gl.glPointSize(5.f);
		gl.glLineWidth(3.f);

		gl.glBegin(GL.GL_LINES);

			double clipAngleR = gClipAngle*Math.PI/180;
			float sinClipAngle = (float)Math.sin(clipAngleR);
			float cosClipAngle = (float)Math.cos(clipAngleR);
			
			// Draw the View Vector starting at the eye (red)
			gl.glColor3f(1, 0, 0);
			gl.glVertex3f(	gViewPosition[0],
							gViewPosition[1]/2,
							gViewPosition[2] );

			gl.glVertex3f(	gViewPosition[0] + 50*sinClipAngle*kGridSpacing,
							gViewPosition[1]/2,
							gViewPosition[2] - 50*cosClipAngle*kGridSpacing );


			double FOV_DIV_2 = gFovX/2*Math.PI/180;
			float sinClipMFov = (float)Math.sin(clipAngleR-FOV_DIV_2);
			float cosClipMFov = (float)Math.cos(clipAngleR-FOV_DIV_2);
			float sinClipPFov = (float)Math.sin(clipAngleR+FOV_DIV_2);
			float cosClipPFov = (float)Math.cos(clipAngleR+FOV_DIV_2);

			// Draw the view frustum (blue)
			gl.glColor3f(0, 0, 1);
			gl.glVertex3f(	gViewPosition[0],
							gViewPosition[1]/2,
							gViewPosition[2] );

			gl.glVertex3f(	gViewPosition[0] + 1000*sinClipMFov*kGridSpacing,
							gViewPosition[1]/2,
							gViewPosition[2] - 1000*cosClipMFov*kGridSpacing );

			gl.glVertex3f(	gViewPosition[0],
							gViewPosition[1]/2,
							gViewPosition[2] );

			gl.glVertex3f(	gViewPosition[0] + 1000*sinClipPFov*kGridSpacing,
							gViewPosition[1]/2,
							gViewPosition[2] - 1000*cosClipPFov*kGridSpacing );

			// Draw the clipping planes behind the eye (yellow)
			float ptEyeX = gViewPosition[0] - kPatchSize*sinClipAngle*kGridSpacing;
			float ptEyeY = gViewPosition[2] + kPatchSize*cosClipAngle*kGridSpacing;

			float ptLeftX = ptEyeX + 100*sinClipMFov*kGridSpacing;
			float ptLeftY = ptEyeY - 100*cosClipMFov*kGridSpacing;

			float ptRightX = ptEyeX + 100*sinClipPFov*kGridSpacing;
			float ptRightY = ptEyeY - 100*cosClipPFov*kGridSpacing;

			gl.glColor3f(1, 1, 0);
			gl.glVertex3f(	ptEyeX,
							gViewPosition[1]/2,
							ptEyeY );

			gl.glVertex3f(	ptLeftX,
							gViewPosition[1]/2,
							ptLeftY);

			gl.glVertex3f(	ptEyeX,
							gViewPosition[1]/2,
							ptEyeY );

			gl.glVertex3f(	ptRightX,
							gViewPosition[1]/2,
							ptRightY);

		gl.glEnd();

		gl.glLineWidth(1.f);
		gl.glColor3f(1, 1, 1);

		setDrawModeContext(gl, gLand.getDrawMode());
	}

	/**
	*  This method loads the terrain data from a disk file.
	**/
	private static HeightMap loadTerrain(int sizeX, int sizeY) throws IOException {
	
		//	The file we want to read in.
		File filePath = new File(kDataFilePath);
		
		//	Map array has one additional row and column so that triangles at edges are
		//	more easily drawn.
		int[][] data = new int[sizeY+1][sizeX+1];
		BufferedInputStream is = new BufferedInputStream(new FileInputStream(filePath));
		
		for (int i=0; i < sizeY; ++i) {
		
			for (int j=0; j < sizeX; ++j)
				data[i][j] = is.read();
				
			//	Copy last column into extra column.
			data[i][sizeX] = data[i][sizeX-1];
		}
		//	Copy last row into extra row.
		System.arraycopy(data[sizeY-1], 0, data[sizeY], 0, sizeX);
		
		is.close();
		
		return new HeightMap2Di(data, -Integer.MAX_VALUE);
	}


	//------------------Required by MouseListener interface--------------------
	private int gStartX = -32768, gStartY = 0;
	
	public void mouseClicked(MouseEvent e) { }
	
	public void mousePressed(MouseEvent e) {
		//	Set a flag indicating that we are starting a rotation.
		gStartX = -32768;
	}
	
	public void mouseReleased(MouseEvent e) { }
	
	public void mouseEntered(MouseEvent e) { }
	
	public void mouseExited(MouseEvent e) { }
	
	//------------------Required by MouseListener interface--------------------
	public void mouseDragged(MouseEvent e) {
		
		if (gCameraMode != FOLLOW_MODE) {
			int mouseX = e.getX();
			int mouseY = e.getY();
			
			// Check for start flag
			if ( gStartX == -32768 ) {
				gStartX = mouseX;
				gStartY = mouseY;
			}
			
			// Find the delta of the mouse
			int dx = mouseX - gStartX;
			int dy = mouseY - gStartY;
			
			// Update the camera rotations
			gCameraRotation[0] += (float) dy * 0.5f;
			gCameraRotation[1] += (float) dx * 0.5f;
	
			// Reset the deltas for the next iteration.
			gStartX = mouseX;
			gStartY = mouseY;
		}
	}
	
	public void mouseMoved(MouseEvent e) { }

	//------------------Required by KeyListener interface--------------------
	public void keyTyped(KeyEvent e) { }
	
	public void keyPressed(KeyEvent e) {
		int key = e.getKeyCode();
		
		if ( key == 'w' || key == 'W')
			doKeyForward();
			
		else if (key == 's' || key == 'S')
			doKeyBackward();
		
		else if (key == 'a' || key == 'A')
			doKeyLeft();
			
		else if (key == 'd' || key == 'D')
			doKeyRight();
	}
	
	public void keyReleased(KeyEvent e) { }
	

	/**
	*  Handles the user pressing the "forward" key.
	**/
	private void doKeyForward() {
	
		switch ( gCameraMode ) {
			default:
			case FOLLOW_MODE:
				break;

			case OBSERVE_MODE:
				gCameraPosition[2] += 5*kGridSpacing;
				break;

			case DRIVE_MODE:
				gViewPosition[0] += (float)(5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_YAW]*Math.PI/180) );
				gViewPosition[2] -= (float)(5*kGridSpacing*Math.cos(gCameraRotation[ROTATE_YAW]*Math.PI/180) );

				gViewPosition[1] = heightAtViewPosition();
				break;

			case FLY_MODE:
				gViewPosition[0] += (float)( 5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_YAW]*Math.PI/180)
										 *Math.cos(gCameraRotation[ROTATE_PITCH]*Math.PI/180) );
				gViewPosition[2] -= (float)( 5*kGridSpacing*Math.cos(gCameraRotation[ROTATE_YAW]*Math.PI/180)
										 *Math.cos( gCameraRotation[ROTATE_PITCH]*Math.PI/180 ) );
				gViewPosition[1] -= (float)( 5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_PITCH]*Math.PI/180) );

				int height = heightAtViewPosition();
				if (gViewPosition[1] < height)
					gViewPosition[1] = height;
				break;
		}

	}
	
	/**
	*  Handles the user pressing the "backward" key.
	**/
	private void doKeyBackward() {
	
		switch ( gCameraMode ) {
			default:
			case FOLLOW_MODE:
				break;
		
			case OBSERVE_MODE:
				gCameraPosition[2] -= 5*kGridSpacing;
				break;

			case DRIVE_MODE:
				gViewPosition[0] -= (float)( 5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_YAW]*Math.PI/180) );
				gViewPosition[2] += (float)( 5*kGridSpacing*Math.cos(gCameraRotation[ROTATE_YAW]*Math.PI/180) );

				gViewPosition[1] = heightAtViewPosition();
				break;

			case FLY_MODE:
				gViewPosition[0] -= (float)( 5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_YAW]*Math.PI/180)
										 *Math.cos(gCameraRotation[ROTATE_PITCH]*Math.PI/180) );
				gViewPosition[2] += (float)( 5*kGridSpacing*Math.cos(gCameraRotation[ROTATE_YAW]*Math.PI/180)
										 *Math.cos(gCameraRotation[ROTATE_PITCH]*Math.PI/180) );
				gViewPosition[1] += (float)( 5*kGridSpacing*Math.sin(gCameraRotation[ROTATE_PITCH]*Math.PI/180) );

				int height = heightAtViewPosition();
				if (gViewPosition[1] < height)
					gViewPosition[1] = height;
				break;
		}
	}

	/**
	*  Returns the height + 4 of the height map directly below the current
	*  view position.
	*  Also limits the view position to be over the height map.
	**/
	private int heightAtViewPosition() {
	
		int posy = (int)(gViewPosition[2]/kGridSpacing);
		int posx = (int)(gViewPosition[0]/kGridSpacing);
		
		if (posy < 0) {
			posy = 0;
			gViewPosition[2] = 0;
		} else if (posy > kMapYSize) {
			posy = kMapYSize;
			gViewPosition[2] = posy*kGridSpacing;
		}
		
		if (posx < 0) {
			posx = 0;
			gViewPosition[0] = 0;
		} else if (posx > kMapXSize) {
			posx = kMapXSize;
			gViewPosition[0] = posx*kGridSpacing;
		}
		
		return (int)((heightMap.get(posx, posy) + 4)*kAltScale);
	}
	
	/**
	*  Handles the user pressing the "rotate left" key.
	**/
	private void doKeyLeft() {
		if ( gCameraMode == OBSERVE_MODE )
			gCameraRotation[1] -= 5.0f;
	}
	
	/**
	*  Handles the user pressing the "rotate right" key.
	**/
	private void doKeyRight() {
		if ( gCameraMode == OBSERVE_MODE )
			gCameraRotation[1] += 5.0f;
	}

}
