/*
*   SoftFPSAnimator -- An Animator with a specified soft frame rate limit.
*
*   Copyright (C) 2004 by Joseph A. Huwaldt
*   All rights reserved.
*   
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2.1 of the License, or (at your option) any later version.
*   
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Lesser General Public License for more details.
*
*   You should have received a copy of the GNU Lesser General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*   Or visit:  http://www.gnu.org/licenses/lgpl.html
**/
package jahuwaldt.gl;

/*
 * Based on Animator.java by Sun Microsystems.  The original Animator.java
 * source contained the following notice:
 *
 * Copyright (c) 2003 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * - Redistribution of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN
 * MIDROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR
 * ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR
 * DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE
 * DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,
 * ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF
 * SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed or intended for use
 * in the design, construction, operation or maintenance of any nuclear
 * facility.
 * 
 * Sun gratefully acknowledges that this software was originally authored
 * and developed by Kenneth Bradley Russell and Christopher John Kline.
 */

import java.awt.EventQueue;
import net.java.games.jogl.impl.SingleThreadedWorkaround;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLCanvas;
import net.java.games.jogl.GLException;

/**
*  <P> This is an Animator which is limited to a maximum frame rate.
*      This prevents the animation from taking 100% CPU available
*      when the excess frame rate is not required.
*      The frame rate is approximate as this is a "soft" approach
*      that can not guarentee a specific or constant frame rate.
*  </P>
*
*  <P> This Animator can be attached to a GLDrawable to drive its
*   display() method in a loop. For efficiency, it sets up the
*   rendering thread for the drawable to be its own internal thread,
*   so it can not be combined with manual repaints of the
*   surface. </P>
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date:  November 23, 2004
*  @version November 27, 2004
**/
public class SoftFPSAnimator {
	private GLDrawable drawable;
	private Runnable runnable;
	private Thread thread;
	private boolean shouldStop;

	/**
	*  The target frames per second to be rendered.
	*/
	private double framesPerSecond=20;

	/**
	*  The target milli-seconds per frame to be rendered.
	**/
	private long mSecPerFrame = (long)((1.0/20.)*1000.0);

	/**
	*  The number of frames that have been counted since the last FPS reset.
	**/
	private long frameCount=0;

	/**
	*  Flag indicating that the frames are currently being counted (or not).
	**/
	private boolean fps_isCounting = true;

	/**
	*  The time, in milliseconds, when the FPS counter was started.
	**/
	private long fps_start = 0;


	/**
	*  Creates a new SoftFPSAnimator for a particular drawable and frame rate.
	*
	*  @param  drawable  A GL drawable object to be animated.
	*  @param  fps       The maximum frame rate to allow in frames per second.
	**/
	public SoftFPSAnimator(GLDrawable drawable, double fps) {
		this.drawable = drawable;
		this.framesPerSecond = fps;
		this.mSecPerFrame = (long)((1.0/fps)*1000.0);
	}

	/** Starts this animator. */
	public synchronized void start() {
		if (thread != null) {
//			throw new GLException("Already started");		//  Why does Animator throw this?
			return;
		}
		if (runnable == null) {
			runnable = new Runnable() {
				public void run() {
					frameCount = 0;
					fps_start = System.currentTimeMillis();

					boolean noException = false;
					try {
						// Try to get OpenGL context optimization since we know we
						// will be rendering this one drawable continually from
						// this thread; make the context current once instead of
						// making it current and freeing it each frame.
						drawable.setRenderingThread(Thread.currentThread());

						// Since setRenderingThread is currently advisory (because
						// of the poor JAWT implementation in the Motif AWT, which
						// performs excessive locking) we also prevent repaint(),
						// which is called from the AWT thread, from having an
						// effect for better multithreading behavior. This call is
						// not strictly necessary, but if end users write their
						// own animation loops which update multiple drawables per
						// tick then it may be necessary to enforce the order of
						// updates.
						drawable.setNoAutoRedrawMode(true);

						while (!shouldStop) {
							long frameStart = System.currentTimeMillis();

							noException = false;
							drawable.display();
							noException = true;

							if (fps_isCounting)
								++frameCount;

							long frameDuration = System.currentTimeMillis() - frameStart;
							if (frameDuration < mSecPerFrame) {
								try {
									long sleepTime = mSecPerFrame - frameDuration;
									Thread.currentThread().sleep(sleepTime, 0);
								} catch (InterruptedException e) {}
							}

						}
					} finally {
						shouldStop = false;
						drawable.setNoAutoRedrawMode(false);
						try {
							// The surface is already unlocked and rendering
							// thread is already null if an exception occurred
							// during display(), so don't disable the rendering
							// thread again.
							if (noException) {
								// Destruction of the underlying GLContext may have
								// disabled the setRenderingThread optimization out
								// from under us
								if (drawable.getRenderingThread() != null) {
									drawable.setRenderingThread(null);
								}
							}
						} finally {
							synchronized (SoftFPSAnimator.this) {
								thread = null;
								SoftFPSAnimator.this.notify();
							}
						}
					}
				}
			};
		}
		thread = new Thread(runnable);
		thread.start();
	}

	/** Stops this animator, blocking until the animation thread has
			finished. */
	public synchronized void stop() {
		shouldStop = true;
		// It's hard to tell whether the thread which calls stop() has
		// dependencies on the Animator's internal thread. Currently we
		// use a couple of heuristics to determine whether we should do
		// the blocking wait().
		if ((Thread.currentThread() == thread) ||
				(SingleThreadedWorkaround.doWorkaround() && EventQueue.isDispatchThread())) {
			return;
		}
		while (shouldStop && thread != null) {
			try {
				wait();
			} catch (InterruptedException ie) {
			}
		}
	}

	/**
	*	Return the approximate frames per second averaged over the number of frames that
	*	have passed since this method was last called or animation began.
	**/
	public synchronized double getFPS() {

		// Stop counting while we are calculating frame rate.
		fps_isCounting = false;

		long fps_end=System.currentTimeMillis();
		long fps_duration = fps_end - fps_start;

		double fps = Double.POSITIVE_INFINITY;
		if (fps_duration > 0) {
			double time = ((double)fps_duration)/1000.0;
			fps	= ((double)frameCount)/time;
		}

		//	Reset FPS measuring variables and start counting again.
		frameCount = 0;
		fps_start=System.currentTimeMillis();
		fps_isCounting = true;

		return fps;
  }

}

